/**
 * Represents view over the UCAN DAG. Can be instantiated with a `root` block
 * and a map of optional proof blocks.
 *
 * @template {UCAN.Capability} Capability
 * @implements {API.Delegation<Capability>}
 */
export class Delegation<Capability extends API.UCAN.Capability<API.UCAN.Ability, `${string}:${string}`>> implements API.Delegation<Capability> {
    /**
     * @param {object} root
     * @param {UCAN.Proof<Capability>} root.cid
     * @param {UCAN.ByteView<UCAN.UCAN<Capability>>} root.bytes
     * @param {UCAN.View<Capability>} [root.data]
     * @param {Map<string, API.Transport.Block<Capability>>} blocks
     */
    constructor({ cid, bytes, data }: {
        cid: UCAN.Proof<Capability>;
        bytes: UCAN.ByteView<UCAN.UCAN<Capability>>;
        data?: API.UCAN.View<Capability> | undefined;
    }, blocks?: Map<string, API.Transport.Block<Capability>>);
    root: {
        cid: API.UCAN.Proof<Capability, number>;
        bytes: API.UCAN.ByteView<API.UCAN.UCAN<Capability>>;
        data: API.UCAN.View<Capability>;
    };
    blocks: Map<string, API.Transport.Block<Capability, number>>;
    get version(): `${number}.${number}.${number}`;
    get signature(): API.UCAN.Signature<Capability>;
    get cid(): API.UCAN.Proof<Capability, number>;
    get bytes(): API.UCAN.ByteView<API.UCAN.UCAN<Capability>>;
    get data(): API.UCAN.View<Capability>;
    /**
     * @returns {IterableIterator<API.Transport.Block<Capability>>}
     */
    export(): IterableIterator<API.Transport.Block<Capability>>;
    get proofs(): API.Proof<Capability>[];
    get issuer(): API.UCAN.DIDView;
    get audience(): API.UCAN.DIDView;
    /**
     * @returns {Capability[]}
     */
    get capabilities(): Capability[];
    /**
     * @returns {number}
     */
    get expiration(): number;
    /**
     * @returns {undefined|number}
     */
    get notBefore(): number | undefined;
    /**
     * @returns {undefined|string}
     */
    get nonce(): string | undefined;
    /**
     * @returns {UCAN.Fact[]}
     */
    get facts(): API.UCAN.Fact[];
}
/**
 * Represents invocation view of the UCAN DAG.
 *
 * @template {API.Capability} Capability
 * @implements {API.InvocationView<Capability>}
 * @extends {Delegation<Capability>}
 */
export class Invocation<Capability extends API.UCAN.Capability<API.UCAN.Ability, `${string}:${string}`>> extends Delegation<Capability> implements API.InvocationView<Capability> {
    get capability(): Capability;
    /**
     * @template {API.InvocationService<Capability>} Service
     * @param {API.Connection<Service>} connection
     * @returns {API.ExecuteInvocation<Capability, Service>}
     */
    execute<Service extends API.InvocationService<Capability, Capability["can"]>>(connection: API.Connection<Service>): API.ExecuteInvocation<Capability, Service, Capability["can"]>;
}
/**
 * @implements {API.Agent}
 */
export class Agent implements API.Agent {
    /**
     * @param {API.UCAN.DID} value
     */
    constructor(value: API.UCAN.DID);
    value: API.UCAN.DID<unknown>;
    did(): API.UCAN.DID<unknown>;
}
/**
 * @template {API.IssuedInvocation[]|API.Invocation[]} Invocations
 * @implements {API.Batch<Invocations>}
 */
export class Batch<Invocations extends API.IssuedInvocation<API.UCAN.Capability<API.UCAN.Ability, `${string}:${string}`>>[] | API.Invocation<API.UCAN.Capability<API.UCAN.Ability, `${string}:${string}`>>[]> implements API.Batch<Invocations> {
    /**
     * @param {Invocations} invocations
     * @param {Map<string, API.UCAN.View>} [delegations]
     */
    constructor(invocations: Invocations, delegations?: Map<string, API.UCAN.View<API.UCAN.Capability<API.UCAN.Ability, `${string}:${string}`>>> | undefined);
    invocations: Invocations;
    delegations: Map<string, API.UCAN.View<API.UCAN.Capability<API.UCAN.Ability, `${string}:${string}`>>>;
}
import * as API from "./api.js";
import * as UCAN from "@ipld/dag-ucan";
import * as DID from "@ipld/dag-ucan/src/did.js";
//# sourceMappingURL=view.d.ts.map